import unittest
import sys


from test_triple_and_curry import TestTripleAndCurry, TestAsyncCurry
from test_effsync import TestEffSync
from test_eff import TestEff


def add_tests_for_class(suite, test_class):
    if sys.version_info < (3, 13):
        suite.addTest(unittest.makeSuite(test_class))
    else:
        suite.addTests(unittest.defaultTestLoader.loadTestsFromTestCase(test_class))


if __name__ == "__main__":
    test_suite = unittest.TestSuite()
    for test_cls in [TestTripleAndCurry, TestAsyncCurry, TestEffSync, TestEff]:
        add_tests_for_class(test_suite, test_cls)
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(test_suite)



import unittest

from mafunca.common.exceptions import MonadError
from mafunca.effsync import EffSync
from mafunca.triple import Right, Left, Nothing


class TestEffSync(unittest.TestCase):

    def test_effect_monad(self):
        eff = EffSync.of(0).map(lambda x: x + 1).bind(lambda x: EffSync(lambda: x + 1))
        self.assertEqual(eff.run(), 2)

        eff = EffSync(lambda: 0).map(lambda x: x + 1).bind(lambda x: EffSync(lambda: x + 1))
        self.assertEqual(eff.run(), 2)

        # working with a good Triple entity
        eff = (
            EffSync.of(0)
            .map(lambda x: Right(x + 1))
            .bind(lambda tm: EffSync(lambda: tm.map(lambda x: x + 1)))
            .bind(lambda tm: EffSync(lambda: tm.map(lambda x: x + 10)))
        )
        self.assertEqual(eff.run().unfold(), 12)
        self.assertTrue(eff.run().is_right)

        # short circuit on bad Triple entities
        eff = (
            EffSync.of(0)
            .map(lambda _: Left('error'))
            .bind(lambda x: EffSync(lambda: x + 1))
            .bind(lambda x: EffSync(lambda: x + 1))
        )
        self.assertEqual(eff.run().unfold(left=lambda x: (x,)), ('error',))

        # short circuit on bad Triple entities
        eff = EffSync.of(0).map(lambda _: Nothing()).map(lambda x: x + 1)
        self.assertEqual(eff.run().get_or_else("There was a Nothing"), "There was a Nothing")
        self.assertTrue(eff.run().is_nothing)

    def test_catch_errors(self):
        def error_raiser():
            raise TypeError

        # catching errors
        eff = (
            EffSync.of(0)
            .map(error_raiser)
            .catch(lambda _: 1)
            .bind(lambda x: EffSync(lambda: x + 1))
            .map(error_raiser)
            .catch(lambda _: EffSync(lambda: -10))
            .bind(lambda x: EffSync(lambda: x + 9))
        )

        self.assertEqual(eff.run(), -1)

        # if there are no errors, the catch method has no effect
        eff = (
            EffSync.of(0)
            .map(lambda x: x + 5)
            .catch(lambda _: 0)     # no effect here
            .bind(lambda x: EffSync(lambda: x + 5))
            .map(lambda x: x + 5)
            .catch(lambda: EffSync(error_raiser))    # no effect here
            .bind(lambda x: EffSync(lambda: x + 5))
        )
        self.assertEqual(eff.run(), 20)

        eff = (
            EffSync(lambda: 0)
            .map(lambda _: error_raiser())
            .catch(lambda _: error_raiser())
            .catch(lambda _: 0)
            .map(lambda x: x + 1)
            .catch(lambda _: EffSync(lambda: -10))   # no effect here
            .bind(lambda x: EffSync(lambda: x + 1))
        )
        self.assertEqual(eff.run(), 2)

        # short circuit on bad Triple entities in catch method
        eff = (
            EffSync.of(0)
            .map(lambda _: error_raiser())
            .catch(lambda _: Nothing())
            .map(lambda x: x + 1)
            .bind(lambda x: EffSync(lambda: x + 1))
        )
        self.assertTrue(eff.run().is_nothing)

    def test_ensure(self):
        def error_raiser():
            raise TypeError

        g = 0

        def side_effect():
            nonlocal g
            g += 10

        with self.assertRaises(TypeError):
            EffSync(lambda: 0).map(lambda _: error_raiser()).ensure(lambda: side_effect()).run()
        self.assertEqual(g, 10)

        eff = EffSync(lambda: 0).map(lambda x: x + 1).ensure(lambda: side_effect()).map(lambda x: x + 1)
        self.assertEqual(eff.run(), 2)
        self.assertEqual(g, 20)

    def test_contract_violation(self):
        async def violated():
            return 1

        with self.assertRaises(MonadError):
            EffSync(violated)

        with self.assertRaises(MonadError):
            EffSync(lambda: 10).map(lambda x: EffSync.of(x)).run()

        with self.assertRaises(MonadError):
            EffSync(lambda: 10).bind(lambda x: x + 1).run()


if __name__ == '__main__':
    unittest.main()



import unittest
import asyncio

from mafunca.triple import Left, Nothing
from mafunca.common.exceptions import MonadError
from mafunca.eff import Eff


class TestEff(unittest.IsolatedAsyncioTestCase):
    async def test_map(self):
        async def plus(x):
            return x + 1

        eff = Eff.of(10).map(lambda x: x + 1)
        self.assertEqual(await eff.run(), 11)

        eff = Eff.of(10).map(plus)
        self.assertEqual(await eff.run(), 11)

    async def test_bind(self):
        async def multiply(x):
            return Eff.of(x * 3)

        eff = Eff.of(5).bind(lambda x: Eff.of(x * 3))
        self.assertEqual(await eff.run(), 15)

        eff = Eff.of(5).bind(multiply)
        self.assertEqual(await eff.run(), 15)

    async def test_short_circuit(self):
        eff = Eff(lambda: Left('err')).map(lambda x: x + 10)
        result = await eff.run()
        self.assertIsInstance(result, Left)
        self.assertEqual(result.unfold(), 'err')

        eff = Eff(lambda: Nothing()).map(lambda x: x + 10)
        result = await eff.run()
        self.assertIsInstance(result, Nothing)

        eff = Eff(lambda: Left("err")).bind(lambda x: Eff.of(x + 1))
        result = await eff.run()
        self.assertIsInstance(result, Left)
        self.assertEqual(result.unfold(), "err")

        eff = Eff(lambda: Nothing()).bind(lambda x: Eff.of(x + 1))
        result = await eff.run()
        self.assertIsInstance(result, Nothing)

    async def test_violation(self):
        eff = Eff.of(1).map(lambda x: Eff.of(x))
        with self.assertRaises(MonadError):
            await eff.run()

        eff = Eff.of(1).bind(lambda x: x + 1)
        with self.assertRaises(MonadError):
            await eff.run()

    async def test_to_thread_violation(self):
        async def violate_sync(x):
            return x + 1

        async def violate_bind_sync():
            return 10

        eff = Eff.of(1)
        with self.assertRaises(MonadError):
            eff.map_to_thread(violate_sync)

        eff = Eff.of(1).bind_to_thread(lambda v: Eff(violate_bind_sync))
        with self.assertRaises(MonadError):
            await eff.run()

        eff = Eff.of(1).map_to_thread(lambda x: Eff.of(x))
        with self.assertRaises(MonadError):
            await eff.run()

        eff = Eff.of(1).bind_to_thread(lambda x: x + 1)
        with self.assertRaises(MonadError):
            await eff.run()

        eff = Eff.of(5).map_to_thread(lambda x: x * 2)
        self.assertEqual(await eff.run(), 10)

        eff = Eff.of(5).bind_to_thread(lambda x: Eff.of(x * 3))
        self.assertEqual(await eff.run(), 15)

    async def test_catch(self):
        async def raiser():
            raise TypeError("error")

        async def catcher(_):
            return 10

        eff = Eff(raiser).catch(lambda e: 10)
        self.assertEqual(await eff.run(), 10)

        eff = Eff(raiser).catch(catcher)
        self.assertEqual(await eff.run(), 10)

        eff = Eff.of(10).map(lambda x: x + 1).catch(lambda e: 0)
        self.assertEqual(await eff.run(), 11)

    async def test_catch_monad_error(self):
        def raises():
            raise MonadError("Eff", "test", "err")

        eff = Eff(raises).catch(lambda e: 42)
        with self.assertRaises(MonadError):
            await eff.run()

    async def test_ensure_sync(self):
        log = []

        def ensure_fn():
            log.append('called')

        def raiser():
            raise TypeError("error")

        wrapped = Eff.of(0).map(lambda _: raiser()).ensure(ensure_fn)
        with self.assertRaises(TypeError):
            await wrapped.run()
        self.assertEqual(log, ['called'])

    async def test_ensure_async(self):
        log = []

        async def eff():
            return 10

        async def ensure_fn():
            log.append('called')

        async def raiser(_):
            raise TypeError("error")

        wrapped = Eff(eff).map(raiser).ensure(ensure_fn)
        with self.assertRaises(TypeError):
            await wrapped.run()
        self.assertEqual(log, ['called'])

    async def test_to_task_requires_async(self):
        eff = Eff.of(2)
        with self.assertRaises(MonadError):
            eff.to_task()

        async def eff_async():
            return 3

        eff2 = Eff(eff_async)
        task = eff2.to_task()
        self.assertTrue(asyncio.isfuture(task))
        result = await task
        self.assertEqual(result, 3)

    async def test_run_delay(self):
        async def fn():
            await asyncio.sleep(2)
            return 1

        eff = Eff(fn)
        with self.assertRaises(TimeoutError):
            await eff.run(delay=1)

        val = await eff.run(delay=3)
        self.assertEqual(val, 1)



from typing import TypeVar, Generic, overload
from collections.abc import Callable

from mafunca.triple import Left, Nothing, TUtils
from mafunca.common.exceptions import MonadError
import mafunca.common.panics as panics


__all__ = ['EffSync']


A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')

L = TypeVar('L')
Exc = TypeVar('Exc', bound=Exception)


class EffSync(Generic[A]):
    """Lazy monad for sync effects.
       It can work with bad 'Triple' entities using the short-circuit principle
    """

    __slots__ = ["effect"]

    def __init__(self, effect: Callable[[], A]):
        panics.on_coroutine(effect, monad_name=self.__class__.__name__, method='__init__')
        self.effect = effect

    @overload
    def map(self: 'EffSync[Left[L]]', fn: Callable[[B], C]) -> 'EffSync[Left[L]]': pass
    @overload
    def map(self: 'EffSync[Nothing]', fn: Callable[[B], C]) -> 'EffSync[Nothing]': pass
    @overload
    def map(self, fn: Callable[[A], B]) -> 'EffSync[B]': pass

    def map(self, fn):
        """
           Applies a sync function that returns a non-EffSync entity.
           :raises MonadError: violation of the contract
        """
        panics.on_coroutine(fn, monad_name=self.__class__.__name__, method='map')

        def new_effect():
            previous = self.effect()
            if TUtils.is_bad(previous):
                return previous
            current = fn(previous)
            panics.on_monadic_result(current, fn=fn, monad=self.__class__, method='map')
            return current
        return EffSync(new_effect)

    @overload
    def bind(self: 'EffSync[Left[L]]', fn: Callable[[B], 'EffSync[C]']) -> 'EffSync[Left[L]]': pass
    @overload
    def bind(self: 'EffSync[Nothing]', fn: Callable[[B], 'EffSync[C]']) -> 'EffSync[Nothing]': pass
    @overload
    def bind(self, fn: Callable[[A], 'EffSync[B]']) -> 'EffSync[B]': pass

    def bind(self, fn):
        """
           Applies a sync function that returns an EffSync entity.
           :raises MonadError: violation of the contract
        """
        panics.on_coroutine(fn, monad_name=self.__class__.__name__, method='bind')

        def new_effect():
            previous = self.effect()
            if TUtils.is_bad(previous):
                return previous
            current_effect = fn(previous)
            panics.on_another_instance(current_effect, fn=fn, monad=self.__class__, method='bind')
            current = current_effect.effect()
            return current
        return EffSync(new_effect)

    @overload
    def catch(self, fn: Callable[[Exc], 'EffSync[B]']) -> 'EffSync[B]': pass
    @overload
    def catch(self, fn: Callable[[Exc], B]) -> 'EffSync[B]': pass

    def catch(self, fn):
        """
           Catch errors(Exception heirs) in all deeper nested functions.
           It can return both EffSync and non-EffSync entities.
           MonadError is not suppressed.
        """
        panics.on_coroutine(fn, monad_name=self.__class__.__name__, method='catch')

        def new_effect():
            try:
                return self.effect()
            except Exception as err:
                if isinstance(err, (MonadError, KeyboardInterrupt)):
                    raise err
                current = fn(err)
                if isinstance(current, self.__class__):
                    return current.effect()
                return current
        return EffSync(new_effect)

    def ensure(self, fn: Callable[[], None]) -> 'EffSync[A]':
        """Guaranteed to execute the function-parameter, similar to try finally"""
        panics.on_coroutine(fn, monad_name=self.__class__.__name__, method='ensure')

        def new_effect():
            try:
                return self.effect()
            finally:
                fn()
        return EffSync(new_effect)

    def run(self) -> A:
        """Starts the chain"""
        return self.effect()

    @staticmethod
    def of(value: A) -> 'EffSync[A]':
        """Wraps a non-EffSync value in the container. No inspections here."""
        return EffSync(lambda: value)

    def __repr__(self):
        return f"EffSync({self.effect})"



from abc import ABC, abstractmethod
from typing import TypeVar, ParamSpec, Generic, Union, overload
from collections.abc import Callable
from functools import wraps

from mafunca.common.exceptions import ImpureMarkError, MonadError
import mafunca.common.panics as panics
from mafunca.curry import Curry


__all__ = [
    'impure',
    'is_impure',
    'Right',
    'Left',
    'Nothing',
    'TUtils',
]


_IMPURE_PROP = 'mafunca_impure'


def impure(fn: Callable):
    """
       Decorator - marks a function as impure(by adding a special attribute)
       to prevent its execution in a triple monad.
       :raises ImpureMarkError: can't mark it for any reason.
    """
    try:
        setattr(fn, _IMPURE_PROP, True)
    except Exception as err:
        raise ImpureMarkError(fn.__name__, str(err))
    return fn


def is_impure(fn: Callable) -> bool:
    """Checking that the function was marked as impure"""
    return getattr(fn, _IMPURE_PROP, False)


R = TypeVar("R")
L = TypeVar("L")

V = TypeVar("V")
W = TypeVar("W")
X = TypeVar("X")


class Triple(ABC):
    """Abstract class for simple monad over the value that is available here and now"""
    @property
    @abstractmethod
    def is_right(self) -> bool:
        pass

    @property
    @abstractmethod
    def is_nothing(self) -> bool:
        pass

    @abstractmethod
    def map(self, fn: Callable[[R], V]):
        pass

    @abstractmethod
    def bind(self, fn: Callable[[R], 'Triple']):
        pass

    @abstractmethod
    def recover_from_left(self, fn: Callable[[L], V]):
        pass

    @abstractmethod
    def recover_from_nothing(self, fn: Callable[[], V]):
        pass

    @abstractmethod
    def unfold(self, *, right: Callable[[R], V], left: Callable[[L], W], nothing: Callable[[], X]):
        pass

    @abstractmethod
    def ap(self, wrapped_val: 'Triple'):
        pass

    @abstractmethod
    def get_or_else(self, alter: V):
        pass

    @abstractmethod
    def __repr__(self):
        pass


def _panic_on_bad_function(fn: Callable, monad: str, method: str, check_impure=True):
    panics.on_coroutine(fn, monad_name=monad, method=method)
    if check_impure and is_impure(fn):
        raise MonadError(monad, method, f"impure function '{fn.__name__}' can not be used")


class _Never:
    """A special marker for a violation of the contract"""
    pass


class Right(Triple, Generic[R]):
    """A branch for a value representing a successful result"""

    __slots__ = ["__value"]

    def __init__(self, value: R):
        self.__value = value

    @property
    def value(self) -> R:
        return self.__value

    @property
    def is_right(self) -> bool:
        return True

    @property
    def is_nothing(self) -> bool:
        return False

    def map(self, fn: Callable[[R], V]) -> 'Right[V]':
        """
           Applies a sync function that returns a non-Triple value.
           :raises MonadError: violation of the contract
        """
        _panic_on_bad_function(fn, monad=self.__class__.__name__, method='map')
        result: V = fn(self.value)
        panics.on_monadic_result(result, fn=fn, monad=Triple, method='map')
        return Right(result)

    @overload
    def bind(self, fn: Callable[[R], 'Right[V]']) -> 'Right[V]': pass
    @overload
    def bind(self, fn: Callable[[R], 'Left[V]']) -> 'Left[V]': pass
    @overload
    def bind(self, fn: Callable[[R], 'Nothing']) -> 'Nothing': pass

    def bind(self, fn):
        """
            Applies a sync function that returns a Triple wrapped value.
            :raises MonadError: violation of the contract
        """
        _panic_on_bad_function(fn, monad=self.__class__.__name__, method='bind')
        return fn(self.value)

    @overload
    def recover_from_left(self, fn: Callable[[L], 'Right[V]']) -> 'Right[R]': pass
    @overload
    def recover_from_left(self, fn: Callable[[L], 'Left[V]']) -> 'Right[R]': pass
    @overload
    def recover_from_left(self, fn: Callable[[L], 'Nothing']) -> 'Right[R]': pass
    @overload
    def recover_from_left(self, fn: Callable[[L], V]) -> 'Right[R]': pass

    def recover_from_left(self, fn):
        """Applies a sync function for recover from error. Always returns itself for this branch."""
        return self

    @overload
    def recover_from_nothing(self, fn: Callable[[], 'Right[V]']) -> 'Right[R]': pass
    @overload
    def recover_from_nothing(self, fn: Callable[[], 'Left[V]']) -> 'Right[R]': pass
    @overload
    def recover_from_nothing(self, fn: Callable[[], 'Nothing']) -> 'Right[R]': pass
    @overload
    def recover_from_nothing(self, fn: Callable[[], V]) -> 'Right[R]': pass

    def recover_from_nothing(self, fn):
        """Applies a sync function for recover from emptiness. Always returns itself for this branch."""
        return self

    def unfold(
            self,
            *,
            right: Callable[[R], V] = lambda v: v,
            left: Callable[[L], W] = lambda w: w,
            nothing: Callable[[], X] = lambda: None
    ) -> V:
        """
            Applies a sync function without wrapping the result. As a rule, it completes the chain.
            :raises MonadError: violation of the contract by 'right'.
        """
        _panic_on_bad_function(right, monad=self.__class__.__name__, method='unfold')
        return right(self.__value)

    @overload
    def ap(self: 'Left[V]', wrapped_val) -> 'Left[V]': pass
    @overload
    def ap(self: 'Nothing', wrapped_val) -> 'Nothing': pass
    @overload
    def ap(self, wrapped_val: 'Left[V]') -> 'Left[V]': pass
    @overload
    def ap(self, wrapped_val: 'Nothing') -> 'Nothing': pass
    @overload
    def ap(self: 'Right[Callable[[V], Right[W]]]', wrapped_val: 'Right[V]') -> 'Right[W]': pass
    @overload
    def ap(self: 'Right[Callable[[V], W]]', wrapped_val: 'Right[V]') -> 'Right[W]': pass
    @overload
    def ap(self: 'Right[Curry[W]]', wrapped_val: 'Right[V]') -> Union['Right[Curry[W]]', V]: pass
    @overload
    def ap(self: 'Right[V]', wrapped_val: 'Right[X]') -> _Never: pass

    def ap(self, wrapped_val):
        """
           Applies a value enclosed in a Triple container to sync function also in the container.
           Combines the logic of map and bind, wrapping simple values in a monad.
           :raises MonadError: violation of the contract.
        """
        _panic_on_bad_function(self.value, monad=self.__class__.__name__, method='ap')
        if not wrapped_val.is_right:
            return wrapped_val
        val = getattr(wrapped_val, 'value')
        result = self.value(val)
        return result if isinstance(result, Triple) else Right(result)

    def get_or_else(self, alter: V) -> R:
        return self.__value

    def __repr__(self):
        return f"Right({self.__value})"


class Left(Triple, Generic[L]):
    """A branch for a value representing an error"""

    __slots__ = ["__value"]

    def __init__(self, value: L):
        self.__value = value

    @property
    def value(self) -> L:
        return self.__value

    @property
    def is_right(self) -> bool:
        return False

    @property
    def is_nothing(self) -> bool:
        return False

    def map(self, fn: Callable[[R], V]) -> 'Left[L]':
        return self

    @overload
    def bind(self, fn: Callable[[R], 'Right[V]']) -> 'Left[L]': pass
    @overload
    def bind(self, fn: Callable[[R], 'Left[V]']) -> 'Left[L]': pass
    @overload
    def bind(self, fn: Callable[[R], 'Nothing']) -> 'Left[L]': pass

    def bind(self, fn):
        return self

    @overload
    def recover_from_left(self, fn: Callable[[L], 'Right[V]']) -> 'Right[V]': pass
    @overload
    def recover_from_left(self, fn: Callable[[L], 'Left[V]']) -> 'Left[V]': pass
    @overload
    def recover_from_left(self, fn: Callable[[L], 'Nothing']) -> 'Nothing': pass
    @overload
    def recover_from_left(self, fn: Callable[[L], V]) -> 'Right[V]': pass

    def recover_from_left(self, fn):
        """
           Applies a sync function for recover from error.
           Combines the logic of map and bind, wrapping simple values in a Triple.
           :raises MonadError: violation of the contract.
        """
        _panic_on_bad_function(fn, monad=self.__class__.__name__, method='recover_from_left')
        result = fn(self.value)
        return result if isinstance(result, Triple) else Right(result)

    @overload
    def recover_from_nothing(self, fn: Callable[[], 'Right[V]']) -> 'Left[L]': pass
    @overload
    def recover_from_nothing(self, fn: Callable[[], 'Left[V]']) -> 'Left[L]': pass
    @overload
    def recover_from_nothing(self, fn: Callable[[], 'Nothing']) -> 'Left[L]': pass
    @overload
    def recover_from_nothing(self, fn: Callable[[], V]) -> 'Left[L]': pass

    def recover_from_nothing(self, fn):
        """Applies a sync function for recover from emptiness. Always returns itself for this branch."""
        return self

    def unfold(
            self,
            *,
            right: Callable[[R], V] = lambda v: v,
            left: Callable[[L], W] = lambda w: w,
            nothing: Callable[[], X] = lambda: None
    ) -> W:
        """
            Applies a sync function without wrapping the result. As a rule, it completes the chain.
            :raises MonadError: violation of the contract by 'left'.
        """
        _panic_on_bad_function(left, monad=self.__class__.__name__, method='unfold')
        return left(self.__value)

    @overload
    def ap(self: 'Left[V]', wrapped_val) -> 'Left[V]': pass
    @overload
    def ap(self: 'Nothing', wrapped_val) -> 'Nothing': pass
    @overload
    def ap(self, wrapped_val: 'Left[V]') -> 'Left[V]': pass
    @overload
    def ap(self, wrapped_val: 'Nothing') -> 'Nothing': pass
    @overload
    def ap(self: 'Right[Callable[[V], Right[W]]]', wrapped_val: 'Right[V]') -> 'Right[W]': pass
    @overload
    def ap(self: 'Right[Callable[[V], W]]', wrapped_val: 'Right[V]') -> 'Right[W]': pass
    @overload
    def ap(self: 'Right[Curry[W]]', wrapped_val: 'Right[V]') -> Union['Right[Curry[W]]', V]: pass
    @overload
    def ap(self: 'Right[V]', wrapped_val: 'Right[X]') -> _Never: pass

    def ap(self, wrapped_val):
        return self

    def get_or_else(self, alter: V) -> V:
        return alter

    def __repr__(self):
        return f"Left({self.__value})"


class Nothing(Triple):
    """A branch for an empty result"""

    __slots__ = []

    @property
    def is_right(self) -> bool:
        return False

    @property
    def is_nothing(self) -> bool:
        return True

    def map(self, fn: Callable[[R], V]) -> 'Nothing':
        return self

    @overload
    def bind(self, fn: Callable[[R], 'Right[V]']) -> 'Nothing': pass
    @overload
    def bind(self, fn: Callable[[R], 'Left[V]']) -> 'Nothing': pass
    @overload
    def bind(self, fn: Callable[[R], 'Nothing']) -> 'Nothing': pass

    def bind(self, fn):
        return self

    @overload
    def recover_from_left(self, fn: Callable[[L], 'Right[V]']) -> 'Nothing': pass
    @overload
    def recover_from_left(self, fn: Callable[[L], 'Left[V]']) -> 'Nothing': pass
    @overload
    def recover_from_left(self, fn: Callable[[L], 'Nothing']) -> 'Nothing': pass
    @overload
    def recover_from_left(self, fn: Callable[[L], V]) -> 'Nothing': pass

    def recover_from_left(self, fn):
        """Applies a sync function for recover from error. Always returns itself for this branch."""
        return self

    @overload
    def recover_from_nothing(self, fn: Callable[[], 'Right[V]']) -> 'Right[V]': pass
    @overload
    def recover_from_nothing(self, fn: Callable[[], 'Left[V]']) -> 'Left[V]': pass
    @overload
    def recover_from_nothing(self, fn: Callable[[], 'Nothing']) -> 'Nothing': pass
    @overload
    def recover_from_nothing(self, fn: Callable[[], V]) -> 'Right[V]': pass

    def recover_from_nothing(self, fn):
        """
           Applies a sync function for recover from emptiness.
           Combines the logic of map and bind, wrapping simple values in a Triple.
           :raises MonadError: violation of the contract.
        """
        _panic_on_bad_function(fn, monad=self.__class__.__name__, method='recover_from_nothing')
        result = fn()
        return result if isinstance(result, Triple) else Right(result)

    def unfold(
            self,
            *,
            right: Callable[[R], V] = lambda v: v,
            left: Callable[[L], W] = lambda w: w,
            nothing: Callable[[], X] = lambda: None
    ) -> V:
        """
            Applies a sync function without wrapping the result. As a rule, it completes the chain.
            :raises MonadError: violation of the contract by 'nothing'.
        """
        _panic_on_bad_function(nothing, monad=self.__class__.__name__, method='unfold')
        return nothing()

    def get_or_else(self, alter: V) -> V:
        return alter

    @overload
    def ap(self: 'Left[V]', wrapped_val) -> 'Left[V]': pass
    @overload
    def ap(self: 'Nothing', wrapped_val) -> 'Nothing': pass
    @overload
    def ap(self, wrapped_val: 'Left[V]') -> 'Left[V]': pass
    @overload
    def ap(self, wrapped_val: 'Nothing') -> 'Nothing': pass
    @overload
    def ap(self: 'Right[Callable[[V], Right[W]]]', wrapped_val: 'Right[V]') -> 'Right[W]': pass
    @overload
    def ap(self: 'Right[Callable[[V], W]]', wrapped_val: 'Right[V]') -> 'Right[W]': pass
    @overload
    def ap(self: 'Right[Curry[W]]', wrapped_val: 'Right[V]') -> Union['Right[Curry[W]]', V]: pass
    @overload
    def ap(self: 'Right[V]', wrapped_val: 'Right[X]') -> _Never: pass

    def ap(self, wrapped_val):
        return self

    def __repr__(self):
        return f"Nothing()"


Params = ParamSpec('Params')


class TUtils:
    """Several useful auxiliary functions - static methods"""

    @staticmethod
    def unit(value: R) -> Right[R]:
        """Wraps a non-Triple value in a Right container"""
        return Right(value)

    @staticmethod
    def from_nullable(value: R, predicate: Callable[[R], bool] = lambda v: v is not None) -> Union[Right[R], Nothing]:
        """Wraps a non-Triple value in a Right container if the predicate returns true, otherwise - Nothing"""
        return Right(value) if predicate(value) else Nothing()

    @staticmethod
    def from_try(fn: Callable[Params, V]) -> Callable[Params, Union[Right[V], Left[Exception]]]:
        """
           Performs a sync function, catching possible errors - heirs of 'Exception'.
           MonadError is not suppressed.
           :raises MonadError: violation of the synchronicity of the function.
        """
        _panic_on_bad_function(fn, monad=TUtils.__name__, method='from_try')

        def from_try_inner(*args: Params.args, **kwargs: Params.kwargs) -> Union[Right[V], Left[Exception]]:
            try:
                result: V = fn(*args, **kwargs)
                return Right(result)
            except Exception as err:
                if isinstance(err, (MonadError, KeyboardInterrupt)):
                    raise err
                return Left(err)

        return wraps(fn)(from_try_inner)

    @staticmethod
    def lift(curried, *wrapped_args: Triple):
        """Applies Triple-wrapped positional arguments to a curried function(not wrapped) through 'ap' method"""
        result = Right(curried)
        for arg in wrapped_args:
            result = result.ap(arg)
        return result

    @staticmethod
    def is_bad(value) -> bool:
        """Check for bad Triple entity"""
        return isinstance(value, Triple) and not value.is_right

    @staticmethod
    def closer(func: Callable[..., R]) -> Callable[..., Union[Left, Nothing, R]]:
        """Sync decorator - if one of the arguments is a "bad" Triple entity, it immediately returns it.
           Otherwise, it calls the function with the passed arguments, automatically unwraps "good" Triple entities.
        """
        def closer_wrapper(*args, **kwargs) -> Union[Left, Nothing, R]:
            for arg in args:
                if TUtils.is_bad(arg):
                    return arg
            for arg in kwargs.values():
                if TUtils.is_bad(arg):
                    return arg
            unwrapped_pos = [getattr(arg, 'value') if isinstance(arg, Triple) else arg for arg in args]
            unwrapped_named = {nm: getattr(v, 'value') if isinstance(v, Triple) else v for nm, v in kwargs.items()}
            return func(*unwrapped_pos, **unwrapped_named)
        return wraps(func)(closer_wrapper)
